use std::borrow::Borrow;
use std::borrow::Cow;
use std::fmt::Debug;
use std::io;
use std::marker::PhantomData;
use std::str::FromStr;
use std::sync::Arc;

use bytemuck::TransparentWrapper;

use crate::Domain;
use crate::DomainDecode;
use crate::Reader;
use crate::DomainEncode;
use crate::Record;
use crate::SignedInteger;
use crate::Symbol;
use crate::Value;
use crate::ValueClass;
use crate::ValueImpl;
use crate::Writer;
use crate::reader::ValueReader;
use crate::shell::read;

#[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
#[repr(transparent)]
pub struct IOValue(pub IOValueImpl);
pub type IOValueImpl = Value<IOValue>;

#[derive(Default)] pub struct IOValueDomainDecode;
#[derive(Default)] pub struct IOValueDomainEncode;

unsafe impl TransparentWrapper<IOValueImpl> for IOValue {}

// pub fn copy_iovalue<V: ValueImpl<'de, D>>(v: &<V::Mapped<V::IOEmbedded> as ValueImpl<'de, D>>::Handle) -> IOValue {
//     Shell::copy::<V::Mapped<V::IOEmbedded>, _, _>(v, &mut |a| Result::<_, ()>::Ok(
//         copy_iovalue::<V>(a.as_ref()).into())).unwrap().into()
// }

impl Domain for IOValue {
    type Decode = IOValueDomainDecode;
    type Encode = IOValueDomainEncode;
}

impl Debug for IOValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        IOValue::peel_ref(self).fmt(f)
    }
}

impl IOValue {
    pub fn new<V: ValueImpl<IOValue> + 'static>(v: V) -> IOValue {
        IOValue(Value::new(v))
    }

    pub fn record(label: IOValue, fields: Vec<IOValue>) -> Self {
        Self::new(Record::new(Value::<IOValue>::from(label),
                              fields.into_iter().map(|f| f.into()).collect()))
    }

    pub fn symbol(s: &str) -> Self {
        Self::new(Symbol::new(s.to_owned()))
    }

    pub fn value(&self) -> &IOValueImpl {
        &self.0
    }
}

impl Borrow<IOValueImpl> for IOValue {
    fn borrow(&self) -> &IOValueImpl {
        &self.0
    }
}

impl FromStr for IOValue {
    type Err = io::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        IOValue::iovalue_from_str(s)
    }
}

impl From<IOValueImpl> for IOValue {
    fn from(h: IOValueImpl) -> Self {
        TransparentWrapper::wrap(h)
    }
}

impl<'r> From<&'r IOValueImpl> for &'r IOValue {
    fn from(h: &'r IOValueImpl) -> Self {
        TransparentWrapper::wrap_ref(h)
    }
}

impl From<IOValue> for IOValueImpl {
    fn from(i: IOValue) -> Self {
        IOValue::peel(i)
    }
}

impl<'r> From<&'r IOValue> for &'r IOValueImpl {
    fn from(i: &'r IOValue) -> Self {
        IOValue::peel_ref(i)
    }
}

impl ValueImpl<IOValue> for IOValue {
    fn write_domain(&self, w: &mut dyn Writer, enc: &mut dyn DomainEncode<IOValue>) -> io::Result<()> {
        self.0.write_domain(w, enc)
    }

    fn owned_value(&self) -> Value<IOValue> {
        IOValue::peel_ref(self).clone()
    }

    fn value_class(&self) -> ValueClass {
        self.value().value_class()
    }

    fn as_boolean(&self) -> Option<bool> {
        self.value().as_boolean()
    }

    fn as_float(&self) -> Option<f32> {
        self.value().as_float()
    }

    fn as_double(&self) -> Option<f64> {
        self.value().as_double()
    }

    fn as_signed_integer(&self) -> Option<Cow<'_, SignedInteger>> {
        self.value().as_signed_integer()
    }

    fn as_string(&self) -> Option<Cow<'_, str>> {
        self.value().as_string()
    }

    fn as_bytestring(&self) -> Option<Cow<'_, [u8]>> {
        self.value().as_bytestring()
    }

    fn as_symbol(&self) -> Option<Cow<'_, str>> {
        self.value().as_symbol()
    }

    fn is_record(&self) -> bool {
        self.value().is_record()
    }

    fn label(&self) -> Value<IOValue> {
        self.value().label().into()
    }

    fn is_sequence(&self) -> bool {
        self.value().is_sequence()
    }

    fn len(&self) -> usize {
        self.value().len()
    }

    fn index(&self, i: usize) -> Value<IOValue> {
        self.value().index(i).into()
    }

    fn iter(&self) -> Box<dyn Iterator<Item = Value<IOValue>> + '_> {
        self.value().iter()
    }

    fn is_set(&self) -> bool {
        self.value().is_set()
    }

    fn has(&self, v: &Value<IOValue>) -> bool {
        self.value().has(v)
    }

    fn is_dictionary(&self) -> bool {
        self.value().is_dictionary()
    }

    fn get(&self, k: &Value<IOValue>) -> Option<Value<IOValue>> {
        self.value().get(k)
    }

    fn entries(&self) -> Box<dyn Iterator<Item = (Value<IOValue>, Value<IOValue>)> + '_> {
        self.value().entries()
    }

    fn as_embedded(&self) -> Option<Cow<'_, IOValue>> {
        self.value().as_embedded()
    }

    fn annotations(&self) -> Option<Cow<'_, [IOValue]>> {
        self.value().annotations()
    }

    fn peeled(self: Arc<Self>) -> Value<IOValue> {
        self.0.clone()
    }
}

impl DomainDecode<IOValue> for IOValueDomainDecode {
    fn decode_embedded<'de, R: Reader<'de> + ?Sized>(
        &mut self,
        r: &mut R,
        read_annotations: bool,
    ) -> io::Result<IOValue> {
        Ok(IOValue::read_iovalue(r, read_annotations)?)
    }
}

impl DomainEncode<IOValue> for IOValueDomainEncode {
    fn encode_embedded(
        &mut self,
        w: &mut dyn Writer,
        d: &IOValue,
    ) -> io::Result<()> {
        (&*d.0).write_domain(w, self)
    }
}

pub struct IOValues<'de, R: Reader<'de>> {
    pub reader: R,
    pub read_annotations: bool,
    phantom: PhantomData<&'de ()>,
}

impl<'de, R: Reader<'de>> IOValues<'de, R> {
    pub fn new(reader: R) -> Self {
        IOValues {
            reader,
            read_annotations: true,
            phantom: PhantomData,
        }
    }

    pub fn read_annotations(mut self, read_annotations: bool) -> Self {
        self.read_annotations = read_annotations;
        self
    }
}

impl<'de, R: Reader<'de>> Iterator for IOValues<'de, R> {
    type Item = io::Result<IOValue>;

    fn next(&mut self) -> Option<Self::Item> {
        match self.reader.peek_class() {
            Err(e) => Some(Err(e)),
            Ok(None) => None,
            Ok(Some(_)) => match IOValue::read_iovalue(&mut self.reader, self.read_annotations) {
                Ok(v) => Some(Ok(v)),
                Err(e) => Some(Err(e.into())),
            }
        }
    }
}

impl ValueReader<IOValue> for IOValue {
    type Impl = IOValue;

    fn read_impl<'de, R: Reader<'de> + ?Sized, Dec: DomainDecode<IOValue>>(
        r: &mut R,
        read_annotations: bool,
        dec: &mut Dec,
    ) -> Result<Self::Impl, crate::error::ReadError> {
        Ok(IOValue(read(r, read_annotations, dec)?))
    }

    fn read_iovalue<'de, R: Reader<'de> + ?Sized>(
        r: &mut R,
        read_annotations: bool,
    ) -> Result<IOValue, crate::error::ReadError> {
        Self::read_impl(r, read_annotations, &mut IOValueDomainDecode::default())
    }
}
