pub mod atom;
pub mod boundary;
pub mod domain;
pub mod error;
pub mod float;
pub mod hex;
pub mod iovalue;
pub mod merge;
pub mod packed;
pub mod reader;
pub mod repr;
pub mod shell;
pub mod signed_integer;
pub mod source;
pub mod text;
pub mod types;
pub mod writer;

#[cfg(feature = "serde")]
pub mod serde;

pub use atom::Atom;
pub use domain::*;
pub use error::Error;
pub use error::ExpectedKind;
pub use iovalue::IOValue;
pub use iovalue::IOValueDomainDecode;
pub use iovalue::IOValueDomainEncode;
pub use iovalue::IOValues;
pub use merge::merge2;
pub use merge::merge;
pub use packed::PackedReader;
pub use packed::PackedWriter;
pub use reader::Reader;
pub use reader::ValueReader;
pub use repr::Value;
pub use repr::ValueImpl;
pub use repr::value_cmp;
pub use repr::value_deepcopy;
pub use repr::value_deepcopy_via;
pub use repr::value_eq;
pub use repr::value_hash;
pub use repr::value_map_embedded;
pub use shell::Annotations;
pub use shell::Bytes;
pub use shell::Embedded;
pub use shell::Map;
pub use shell::Record;
pub use shell::Set;
pub use shell::Symbol;
pub use shell::TreeReader;
pub use signed_integer::SignedInteger;
pub use source::BinarySource;
pub use source::BytesBinarySource;
pub use source::IOBinarySource;
pub use text::TextReader;
pub use text::TextWriter;
pub use types::AtomClass;
pub use types::CompoundClass;
pub use types::ValueClass;
pub use writer::Writer;
pub use writer::write_value;

#[cfg(test)]
mod demo {
    use crate::*;

    // fn getit<'a>(d: &'a dyn ValueImpl<NoValue>, k: &str) -> Option<&'a dyn ValueImpl<NoValue>> {
    //     d.get(&k)
    // }

    #[test] fn a() {
        let l: IOValue = "label".parse().unwrap();
        let r = IOValue::record(l.clone(), vec![IOValue::new(1), IOValue::new(2), IOValue::new(3)]);
        let r2 = IOValue::record(l, vec![IOValue::new(1), IOValue::new(2), IOValue::new(4)]);

        let mut v: Map<IOValue, IOValue> = Map::new();
        v.insert("\"abc\"".parse().unwrap(), "def".parse().unwrap());
        v.insert("abc".parse().unwrap(), "DEF".parse().unwrap());
        v.insert(IOValue::new(123), "xyz".parse().unwrap());
        v.insert(IOValue::new(vec![1, 2, 3]), "{a: 1, b: 2}".parse().unwrap());
        v.insert(r2, "bbb".parse().unwrap());
        v.insert(r, "<foo bar zot>".parse().unwrap());
        let w: &dyn ValueImpl<IOValue> = &v;
        println!("GETw abc {:?}", w.get(&IOValue::new("abc").into()));
        println!("GETw 123 {:?}", w.get(&IOValue::new(123).into()));
        println!("GETw qqq {:?}", w.get(&IOValue::new("qqq").into()));
        println!("GETv abc {:?}", v.get(&IOValue::new("abc")));
        println!("GETv 123 {:?}", v.get(&IOValue::new(123)));
        println!("GETv qqq {:?}", v.get(&IOValue::new("qqq")));
        for (kk, vv) in w.entries() {
            println!("{:#?} ==> {:#?}", kk, vv);
        }

        // {
        //     use std::io::BufRead;
        //     for line in std::io::stdin().lock().lines() {
        //         let line = line.unwrap();
        //         let key = line.parse::<PlainValue<NoValue>>().unwrap();
        //         let val = w.get(&key);
        //         println!("{:?} == {:?} ==> {:?} == {:?}", line, &key, val, getit(&v, &line));
        //     }
        // }
    }

    #[test] fn value_size() {
        println!("Value size {}", std::mem::size_of::<Value<IOValue>>());
        println!("&dyn ValueImpl size {}", std::mem::size_of::<&dyn ValueImpl<IOValue>>());
        println!("Box dyn ValueImpl size {}", std::mem::size_of::<Box<dyn ValueImpl<IOValue>>>());
    }
}

#[cfg(test)]
mod test_domain {
    use std::io;
    use crate::*;

    #[derive(Debug, Hash, Clone, Ord, PartialEq, Eq, PartialOrd)]
    pub enum Dom {
        One,
        Two,
    }

    impl Domain for Dom {
        type Decode = DebugDomainCodec;
        type Encode = DebugDomainCodec;
    }

    impl std::str::FromStr for Dom {
        type Err = io::Error;
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            match s {
                "One" => Ok(Dom::One),
                "Two" => Ok(Dom::Two),
                _ => Err(io::Error::new(io::ErrorKind::Other, "cannot parse preserves test domain")),
            }
        }
    }

    struct DomCodec;

    impl DomainDecode<Dom> for DomCodec {
        fn decode_embedded<'de, R: Reader<'de> + ?Sized>(
            &mut self,
            r: &mut R,
            _read_annotations: bool,
        ) -> io::Result<Dom> {
            let v = IOValue::read_iovalue(r, false)?;
            if v.value().as_bytestring().is_some() {
                Ok(Dom::One)
            } else {
                Ok(Dom::Two)
            }
        }
    }

    impl DomainEncode<Dom> for DomCodec {
        fn encode_embedded(
            &mut self,
            w: &mut dyn Writer,
            d: &Dom,
        ) -> io::Result<()> {
            match d {
                Dom::One => IOValue::new(Bytes::new(vec![255, 255, 255, 255])).write(w),
                Dom::Two => IOValue::new(Symbol::new(format!("Dom::{:?}", d))).write(w),
            }
        }
    }

    fn dom_as_preserves(v: &Dom) -> io::Result<IOValue> {
        Ok(match v {
            Dom::One => IOValue::new(Bytes::new(vec![255, 255, 255, 255])),
            Dom::Two => IOValue::new(Symbol::new(format!("Dom::{:?}", v))),
        })
    }

    #[test] fn test_one() {
        let v = Value::new(
            vec![Value::new(SignedInteger::from(1)),
                 Value::new(Embedded::new(Dom::One)),
                 Value::new(SignedInteger::from(2))]);
        assert_eq!(PackedWriter::encode(&value_deepcopy_via(&v, &mut dom_as_preserves).unwrap()).unwrap(),
                   [0xb5, 0x91, 0xb2, 0x04, 255, 255, 255, 255, 0x92, 0x84]);
    }

    #[test] fn test_two() {
        let v = Value::new(
            vec![Value::new(SignedInteger::from(1)),
                 Value::new(Embedded::new(Dom::Two)),
                 Value::new(SignedInteger::from(2))]);
        assert_eq!(PackedWriter::encode(&value_deepcopy_via(&v, &mut dom_as_preserves).unwrap()).unwrap(),
                   [0xb5, 0x91, 0xb3, 0x08, 68, 111, 109, 58, 58, 84, 119, 111, 0x92, 0x84]);
    }
}
