use std::borrow::Cow;
use std::convert::TryFrom;
use std::io;

use crate::Atom;
use crate::BinarySource;
use crate::BytesBinarySource;
use crate::CompoundClass;
use crate::DefaultDomainCodec;
use crate::Domain;
use crate::DomainDecode;
use crate::IOValue;
use crate::SignedInteger;
use crate::Value;
use crate::ValueClass;
use crate::ValueImpl;
use crate::boundary as B;
use crate::error::Error;
use crate::error::ExpectedKind;
use crate::error::ReadError;
use crate::error::io_eof;

pub type ReaderResult<T> = std::result::Result<T, Error>;

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum NextToken {
    Annotation,
    Value(ValueClass),
}

pub trait ValueReader<D: Domain> {
    type Impl: ValueImpl<D> + 'static;

    fn read_impl<'de, R: Reader<'de> + ?Sized, Dec: DomainDecode<D>>(
        r: &mut R,
        read_annotations: bool,
        dec: &mut Dec,
    ) -> Result<Self::Impl, ReadError>;

    fn read_iovalue<'de, R: Reader<'de> + ?Sized>(
        r: &mut R,
        read_annotations: bool,
    ) -> Result<IOValue, ReadError>;

    //---------------------------------------------------------------------------

    fn from_str_domain<Dec: DomainDecode<D>>(
        s: &str,
        read_annotations: bool,
        dec: &mut Dec,
    ) -> io::Result<Value<D>> {
        Ok(Self::read_domain(&mut BytesBinarySource::new(s.as_bytes()).text(), read_annotations, dec)?)
    }

    fn from_str(
        s: &str,
        read_annotations: bool,
    ) -> io::Result<Value<D>> {
        Self::from_str_domain(s, read_annotations, &mut DefaultDomainCodec)
    }

    fn iovalue_from_str(s: &str) -> io::Result<IOValue> {
        Ok(Self::read_iovalue(&mut BytesBinarySource::new(s.as_bytes()).text(), true)?)
    }

    fn from_bytes_domain<Dec: DomainDecode<D>>(
        bs: &[u8],
        read_annotations: bool,
        dec: &mut Dec,
    ) -> io::Result<Value<D>> {
        Ok(Self::read_domain(&mut BytesBinarySource::new(bs).packed(), read_annotations, dec)?)
    }

    fn from_bytes(
        bs: &[u8],
        read_annotations: bool,
    ) -> io::Result<Value<D>> {
        Self::from_bytes_domain(bs, read_annotations, &mut DefaultDomainCodec)
    }

    fn iovalue_from_bytes(bs: &[u8]) -> io::Result<IOValue> {
        Ok(Self::read_iovalue(&mut BytesBinarySource::new(bs).packed(), true)?)
    }

    fn read_domain<'de, R: Reader<'de> + ?Sized, Dec: DomainDecode<D>>(
        r: &mut R,
        read_annotations: bool,
        dec: &mut Dec,
    ) -> Result<Value<D>, ReadError> {
        Ok(Value::new(Self::read_impl(r, read_annotations, dec)?))
    }

    fn read<'de, R: Reader<'de> + ?Sized>(
        r: &mut R,
        read_annotations: bool,
    ) -> Result<Value<D>, ReadError> {
        Self::read_domain(r, read_annotations, &mut D::Decode::default())
    }

    fn gather_annotations<'de, R: Reader<'de> + ?Sized>(
        r: &mut R,
    ) -> io::Result<Option<(Vec<IOValue>, ValueClass)>> {
        let mut anns = Vec::new();
        loop {
            match r.peek_class()? {
                None => return Ok(None),
                Some(NextToken::Value(v)) => return Ok(Some((anns, v))),
                Some(NextToken::Annotation) => {
                    r.open_annotation()?;
                    anns.push(Self::read_iovalue(r, true)?);
                    r.close_annotation()?;
                }
            }
        }
    }
}

pub trait Reader<'de> {
    fn peek_class(&mut self) -> io::Result<Option<NextToken>>;

    fn next_atom(&mut self) -> ReaderResult<Atom<'de>>;

    fn boundary(&mut self, b: &B::Type) -> ReaderResult<()>;

    fn open_record(&mut self) -> ReaderResult<()>;
    fn open_sequence(&mut self) -> ReaderResult<()>;
    fn open_set(&mut self) -> ReaderResult<()>;
    fn open_dictionary(&mut self) -> ReaderResult<()>;

    // Answers true for closed, false for more.
    // Implies a b.shift of None if closed or of Some(i) if not closed, plus a .boundary.
    fn close_compound(&mut self, b: &mut B::Type, i: &B::Item) -> ReaderResult<bool>;

    fn open_embedded(&mut self) -> ReaderResult<()>;
    fn close_embedded(&mut self) -> ReaderResult<()>;

    fn open_annotation(&mut self) -> ReaderResult<()>;
    fn close_annotation(&mut self) -> ReaderResult<()>;

    fn mark(&mut self) -> io::Result<usize>;
    fn restore(&mut self, mark: usize) -> io::Result<()>;

    //---------------------------------------------------------------------------

    fn skip_atom(&mut self) -> io::Result<()> {
        let _ = self.next_atom()?;
        Ok(())
    }

    fn skip_annotations(&mut self) -> io::Result<Option<ValueClass>> {
        loop {
            match self.peek_class()? {
                None => return Ok(None),
                Some(NextToken::Value(v)) => return Ok(Some(v)),
                Some(NextToken::Annotation) => {
                    self.open_annotation()?;
                    self.skip_value()?;
                    self.close_annotation()?;
                }
            }
        }
    }

    fn skip_value(&mut self) -> io::Result<()> {
        // TODO efficient skipping in specific impls of this trait
        match self.skip_annotations()?.ok_or_else(io_eof)? {
            ValueClass::Atomic(_) => self.skip_atom(),
            ValueClass::Embedded => {
                self.open_embedded()?;
                self.skip_value()?;
                self.close_embedded()?;
                Ok(())
            }
            ValueClass::Compound(CompoundClass::Record) => {
                self.open_record()?;
                let mut b = B::start(B::Item::RecordLabel);
                self.boundary(&b)?;
                self.skip_value()?;
                while !self.close_compound(&mut b, &B::Item::RecordField)? {
                    self.skip_value()?;
                }
                Ok(())
            }
            ValueClass::Compound(CompoundClass::Sequence) => {
                self.open_sequence()?;
                let mut b = B::Type::default();
                while !self.close_compound(&mut b, &B::Item::SequenceValue)? {
                    self.skip_value()?;
                }
                Ok(())
            }
            ValueClass::Compound(CompoundClass::Set) => {
                self.open_set()?;
                let mut b = B::Type::default();
                while !self.close_compound(&mut b, &B::Item::SetValue)? {
                    self.skip_value()?;
                }
                Ok(())
            }
            ValueClass::Compound(CompoundClass::Dictionary) => {
                self.open_dictionary()?;
                let mut b = B::Type::default();
                while !self.close_compound(&mut b, &B::Item::DictionaryKey)? {
                    self.skip_value()?;
                    b.shift(Some(B::Item::DictionaryValue));
                    self.boundary(&b)?;
                    self.skip_value()?;
                }
                Ok(())
            }
        }
    }

    fn next_boolean(&mut self) -> ReaderResult<bool> { Ok(self.next_atom()?.try_into()?)}
    fn next_float(&mut self) -> ReaderResult<f32> { Ok(self.next_atom()?.try_into()?)}
    fn next_double(&mut self) -> ReaderResult<f64> { Ok(self.next_atom()?.try_into()?)}
    fn next_signedinteger(&mut self) -> ReaderResult<Cow<'de, SignedInteger>> { Ok(self.next_atom()?.try_into()?) }
    fn next_str(&mut self) -> ReaderResult<Cow<'de, str>> { Ok(self.next_atom()?.try_into()?) }
    fn next_bytestring(&mut self) -> ReaderResult<Cow<'de, [u8]>> { Ok(self.next_atom()?.try_into()?) }
    fn next_symbol(&mut self) -> ReaderResult<Cow<'de, str>> { Ok(self.next_atom()?.try_into_symbol()?) }

    fn next_i8(&mut self) -> ReaderResult<i8> { Ok(i8::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_u8(&mut self) -> ReaderResult<u8> { Ok(u8::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_i16(&mut self) -> ReaderResult<i16> { Ok(i16::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_u16(&mut self) -> ReaderResult<u16> { Ok(u16::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_i32(&mut self) -> ReaderResult<i32> { Ok(i32::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_u32(&mut self) -> ReaderResult<u32> { Ok(u32::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_i64(&mut self) -> ReaderResult<i64> { Ok(i64::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_u64(&mut self) -> ReaderResult<u64> { Ok(u64::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_i128(&mut self) -> ReaderResult<i128> { Ok(i128::try_from(self.next_signedinteger()?.as_ref())?) }
    fn next_u128(&mut self) -> ReaderResult<u128> { Ok(u128::try_from(self.next_signedinteger()?.as_ref())?) }

    fn open_simple_record(&mut self, name: &str) -> ReaderResult<B::Type>
    {
        self.open_record()?;
        let b = B::start(B::Item::RecordLabel);
        self.boundary(&b)?;
        let label: &str = &self.next_symbol()?;
        if label == name {
            Ok(b)
        } else {
            Err(Error::Expected(ExpectedKind::SimpleRecord(name.to_owned())))
        }
    }

    fn ensure_more_expected(&mut self, b: &mut B::Type, i: &B::Item) -> ReaderResult<()> {
        if !self.close_compound(b, i)? {
            Ok(())
        } else {
            Err(Error::MissingItem)
        }
    }

    fn ensure_complete(&mut self, mut b: B::Type, i: &B::Item) -> ReaderResult<()> {
        if !self.close_compound(&mut b, i)? {
            Err(Error::MissingCloseDelimiter)
        } else {
            Ok(())
        }
    }

    fn specialized(&mut self) -> Option<(&str, &mut dyn BinarySource<'de>)> { None }
}
