use crate::{Value, IOValue, ValueImpl, ExpectedKind};
use crate::serde::error::Error;
use serde::Deserialize;
use serde::de::{Visitor, SeqAccess, MapAccess, EnumAccess, VariantAccess, DeserializeSeed};
use std::iter::Iterator;

pub type Result<T> = std::result::Result<T, Error>;

pub struct Deserializer {
    input: IOValue,
}

pub fn from_value<'de, T>(v: &IOValue) -> Result<T> where T: Deserialize<'de>
{
    let mut de = Deserializer::from_value(v);
    let t = T::deserialize(&mut de)?;
    Ok(t)
}

impl Deserializer {
    pub fn from_value(v: &IOValue) -> Self {
        Deserializer { input: v.clone() }
    }
}

impl<'de, 'a> serde::de::Deserializer<'de> for &'a mut Deserializer
{
    type Error = Error;

    fn deserialize_any<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        match self.input.value_class() {
            crate::ValueClass::Atomic(a) => match a {
                crate::AtomClass::Boolean => visitor.visit_bool(self.input.as_boolean().unwrap()),
                crate::AtomClass::Float => visitor.visit_f32(self.input.as_float().unwrap()),
                crate::AtomClass::Double => visitor.visit_f64(self.input.as_double().unwrap()),
                crate::AtomClass::SignedInteger => {
                    let v = self.input.as_signed_integer().unwrap().into_owned();
                    match i64::try_from(&v) {
                        Ok(i) => visitor.visit_i64(i),
                        Err(_) => match u64::try_from(&v) {
                            Ok(u) => visitor.visit_u64(u),
                            Err(_) => Err(crate::Error::NumberOutOfRange((&v).into()))?,
                        }
                    }
                }
                crate::AtomClass::String => visitor.visit_str(self.input.as_string().unwrap().as_ref()),
                crate::AtomClass::ByteString => self.deserialize_bytes(visitor),
                crate::AtomClass::Symbol => Err(Error::CannotDeserializeAny),
            }
            crate::ValueClass::Compound(c) => match c {
                crate::CompoundClass::Record => match self.input.label().as_symbol() {
                    Some(s) => match s.as_ref() {
                        "tuple" => if self.input.len() == 0 {
                            self.deserialize_unit(visitor)
                        } else {
                            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
                        },
                        "UnicodeScalar" => self.deserialize_char(visitor),
                        "None" | "Some" => self.deserialize_option(visitor),
                        _ => Err(Error::CannotDeserializeAny),
                    }
                    _ => Err(Error::CannotDeserializeAny),
                }
                crate::CompoundClass::Sequence => visitor.visit_seq(VecSeq::new(self, self.input.clone().iter())),
                crate::CompoundClass::Set => Err(Error::CannotDeserializeAny),
                crate::CompoundClass::Dictionary => visitor.visit_map(DictMap::new(self, self.input.clone().entries())),
            }
            crate::ValueClass::Embedded => Err(Error::CannotDeserializeAny),
        }
    }

    fn deserialize_bool<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_bool(self.input.to_boolean()?)
    }

    fn deserialize_i8<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_i8(self.input.to_i8()??)
    }

    fn deserialize_i16<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_i16(self.input.to_i16()??)
    }

    fn deserialize_i32<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_i32(self.input.to_i32()??)
    }

    fn deserialize_i64<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_i64(self.input.to_i64()??)
    }

    fn deserialize_u8<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_u8(self.input.to_u8()??)
    }

    fn deserialize_u16<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_u16(self.input.to_u16()??)
    }

    fn deserialize_u32<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_u32(self.input.to_u32()??)
    }

    fn deserialize_u64<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_u64(self.input.to_u64()??)
    }

    fn deserialize_f32<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        match self.input.as_float() {
            Some(f) => visitor.visit_f32(f),
            None => visitor.visit_f32(self.input.to_double()? as f32),
        }
    }

    fn deserialize_f64<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        match self.input.as_double() {
            Some(d) => visitor.visit_f64(d),
            None => visitor.visit_f64(self.input.to_float()? as f64),
        }
    }

    fn deserialize_char<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record("UnicodeScalar", Some(1)) {
            if let Some(Ok(c)) = self.input.index(0).as_u32() {
                return visitor.visit_char(char::from_u32(c).ok_or(Error::InvalidUnicodeScalar(c))?)
            }
        }
        Err(Error::ExpectedUnicodeScalar)
    }

    fn deserialize_str<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_str(self.input.to_str()?.as_ref())
    }

    fn deserialize_string<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        self.deserialize_str(visitor)
    }

    fn deserialize_bytes<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_bytes(self.input.to_bytestring()?.as_ref())
    }

    fn deserialize_byte_buf<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_byte_buf(self.input.to_bytestring()?.into_owned())
    }

    fn deserialize_option<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        let i = self.input.clone();
        if i.is_record() {
            if let Some(s) = i.label().as_symbol() {
                match s.as_ref() {
                    "None" => if let None = i.iter().next() {
                        return visitor.visit_none();
                    },
                    "Some" => if let Some(v) = i.iter().next() {
                        self.input = v.into();
                        return visitor.visit_some(self);
                    },
                    _ => (),
                }
            }
        }
        Err(Error::ExpectedOption)
    }

    fn deserialize_unit<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record("tuple", Some(0)) {
            visitor.visit_unit()
        } else {
            Err(ExpectedKind::SimpleRecord("tuple".to_owned()))?
        }
    }

    fn deserialize_unit_struct<V>(self, name: &'static str, visitor: V)
                                  -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record(name, Some(0)) {
            visitor.visit_unit()
        } else {
            Err(ExpectedKind::SimpleRecord(name.to_owned()))?
        }
    }

    fn deserialize_newtype_struct<V>(self, name: &'static str, visitor: V)
                                     -> Result<V::Value> where V: Visitor<'de>
    {
        match super::magic::transmit_input_value(name, || Ok(self.input.clone()))? {
            Some(v) => visitor.visit_u64(v),
            None => if self.input.is_simple_record(name, Some(1)) {
                self.input = self.input.index(0).into();
                visitor.visit_newtype_struct(self)
            } else {
                Err(ExpectedKind::SimpleRecord(name.to_owned()))?
            }
        }
    }

    fn deserialize_seq<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de> {
        if self.input.is_sequence() {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else if self.input.is_set() {
            // Hack around serde's model: Deserialize *sets* as
            // sequences, too, and reconstruct them as Rust Sets
            // on the visitor side.
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::Sequence)?
        }
    }

    fn deserialize_tuple<V>(self, len: usize, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record("tuple", Some(len)) {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::SimpleRecord("tuple".to_owned()))?
        }
    }

    fn deserialize_tuple_struct<V>(self, name: &'static str, len: usize, visitor: V)
                                   -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record(name, Some(len)) {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::SimpleRecord(name.to_owned()))?
        }
    }

    fn deserialize_map<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de> {
        if self.input.is_dictionary() {
            visitor.visit_map(DictMap::new(self, self.input.clone().entries()))
        } else {
            Err(ExpectedKind::Dictionary)?
        }
    }

    fn deserialize_struct<V>(self,
                             name: &'static str,
                             fields: &'static [&'static str],
                             visitor: V)
                             -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_simple_record(name, Some(fields.len())) {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::SimpleRecord(name.to_owned()))?
        }
    }

    fn deserialize_enum<V>(self,
                           _name: &'static str,
                           _variants: &'static [&'static str],
                           visitor: V)
                           -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_enum(self)
    }

    fn deserialize_identifier<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_str(&self.input.value().to_symbol()?)
    }

    fn deserialize_ignored_any<V>(self, visitor: V) -> Result<V::Value> where V: Visitor<'de>
    {
        visitor.visit_none()
    }
}

pub struct VecSeq<'a, I: Iterator<Item = Value<IOValue>>> {
    iter: I,
    de: &'a mut Deserializer,
}

impl<'de, 'a, I: Iterator<Item = Value<IOValue>>> VecSeq<'a, I> {
    fn new(de: &'a mut Deserializer, iter: I) -> Self {
        VecSeq { iter, de }
    }
}

impl<'de, 'a, I: Iterator<Item = Value<IOValue>>> SeqAccess<'de> for VecSeq<'a, I> {
    type Error = Error;

    fn next_element_seed<T>(&mut self, seed: T) ->
        Result<Option<T::Value>>
    where
        T: DeserializeSeed<'de>
    {
        match self.iter.next() {
            None => Ok(None),
            Some(v) => {
                self.de.input = v.into();
                Ok(Some(seed.deserialize(&mut *self.de)?))
            }
        }
    }
}

pub struct DictMap<'a> {
    pending: Option<IOValue>,
    iter: Box<dyn Iterator<Item = (Value<IOValue>, Value<IOValue>)> + 'a>,
    de: &'a mut Deserializer,
}

impl<'a> DictMap<'a> {
    fn new(
        de: &'a mut Deserializer,
        iter: Box<dyn Iterator<Item = (Value<IOValue>, Value<IOValue>)> + 'a>,
    ) -> Self {
        DictMap{ pending: None, iter, de }
    }
}

impl<'a, 'de> MapAccess<'de> for DictMap<'a> {
    type Error = Error;

    fn next_key_seed<K>(
        &mut self,
        seed: K,
    ) -> Result<Option<K::Value>> where K: DeserializeSeed<'de> {
        match self.iter.next() {
            None => Ok(None),
            Some((k, v)) => {
                self.pending = Some(v.into());
                self.de.input = k.into();
                Ok(Some(seed.deserialize(&mut *self.de)?))
            }
        }
    }

    fn next_value_seed<V>(&mut self, seed: V) -> Result<V::Value> where V: DeserializeSeed<'de> {
        let v = self.pending.take().unwrap();
        self.de.input = v;
        Ok(seed.deserialize(&mut *self.de)?)
    }
}

impl<'a, 'de> EnumAccess<'de> for &'a mut Deserializer {
    type Error = Error;
    type Variant = Self;

    fn variant_seed<V>(self, seed: V)
                       -> Result<(V::Value, Self::Variant)> where V: DeserializeSeed<'de>
    {
        let v = self.input.clone();
        self.input = v.label().into();
        let variant = seed.deserialize(&mut *self)?;
        self.input = v;
        Ok((variant, self))
    }
}

impl<'a, 'de> VariantAccess<'de> for &'a mut Deserializer {
    type Error = Error;

    fn unit_variant(self) -> Result<()> {
        Ok(())
    }

    fn newtype_variant_seed<T>(self, seed: T) -> Result<T::Value> where T: DeserializeSeed<'de> {
        let i = self.input.clone();
        if i.is_record() {
            if let Some(v) = i.iter().next() {
                self.input = v.into();
                return seed.deserialize(&mut *self);
            }
        }
        Err(ExpectedKind::Record)?
    }

    fn tuple_variant<V>(self, _len: usize, visitor: V) -> Result<V::Value> where V: Visitor<'de> {
        if self.input.is_record() {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::Record)?
        }
    }

    fn struct_variant<V>(self, _fields: &'static [&'static str], visitor: V)
                         -> Result<V::Value> where V: Visitor<'de>
    {
        if self.input.is_record() {
            visitor.visit_seq(VecSeq::new(self, self.input.clone().iter()))
        } else {
            Err(ExpectedKind::Record)?
        }
    }
}
